/**
 * ftq.c : Fixed Time Quantum microbenchmark
 *
 * Written by Matthew Sottile (matt@lanl.gov)
 *
 * This is a complete rewrite of the original tscbase code by
 * Ron and Matt, in order to use a better set of portable timers,
 * and more flexible argument handling and parameterization.
 *
 * Licensed under the terms of the GNU Public Licence.  See LICENCE_GPL 
 * for details.
 */
#include <u.h>
#include <libc.h>

/**
 * macros and defines
 */
#define MAX_SAMPLES    2000000
#define DEFAULT_COUNT  10000
#define DEFAULT_BITS   20
#define MAX_BITS       30
#define MIN_BITS       3
#define CORE63

#ifdef CORE15
#define MULTIITER
#define ITERCOUNT      8
#endif
#ifdef CORE31
#define MULTIITER
#define ITERCOUNT      16
#endif
#ifdef CORE63
#define MULTIITER
#define ITERCOUNT      32
#endif

/**
 * global variables
 */

/* samples: each sample has a timestamp and a work count. */
static unsigned long long samples[MAX_SAMPLES*2];
static uvlong now, last, endinterval;
static unsigned long done;

/**
 * usage()
 */
void usage(char *av0) {
  fprint(2,"usage: %s [-n samples] [-i bits] [-h] [-o outname] [-s]\n",
	  av0);
  exits("failure");
}

/**
 * main()
 */
void main(int argc, char **argv) {
  /* local variables */
  unsigned long numsamples = DEFAULT_COUNT;
  unsigned long long interval_length;
  int interval_bits = DEFAULT_BITS;
  unsigned long long count;
  char fname_times[1024], fname_counts[1024];
  int i;
  int fp;
  int use_stdout = 0;
  char* tmp;
#ifdef MULTIITER
  register int k;
#endif


#ifdef _WITH_TAU_
  TAU_PROFILE("main()", "int (int, char **)", TAU_DEFAULT);
  TAU_PROFILE_TIMER(mtimer,"main","void (void)", TAU_USER);
  TAU_PROFILE_TIMER(t1,"coretimer","void (void)", TAU_USER);
  TAU_PROFILE_TIMER(looptimer,"looptimer","void (void)", TAU_USER);
  TAU_PROFILE_TIMER(datasave,"datasave","void (void)", TAU_USER);
  TAU_PROFILE_TIMER(io,"io","void (void)", TAU_USER);
  TAU_PROFILE_INIT(argc, argv);
  TAU_PROFILE_SET_NODE(0);
  TAU_PROFILE_START(mtimer);
#endif /* _WITH_TAU_ */

  /* default filenames */
  sprint(fname_times,"ftq_times.dat");
  sprint(fname_counts,"ftq_counts.dat");

	ARGBEGIN{
    case 's':
      use_stdout = 1;
      break;
    case 'o':
	  tmp = EARGF(usage(argv[0]));
      sprint(fname_times,"%s_times.dat",tmp);
      sprint(fname_counts,"%s_counts.dat",tmp);  
      break;
    case 'i':
      interval_bits = atoi(EARGF(usage(argv[0])));
      break;
    case 'n':
      numsamples = atoi(EARGF(usage(argv[0])));
      break;
    case 'h':
    default:
      usage(argv[0]);
      break;
	}ARGEND

	fp = create(fname_times,OWRITE,0666);

	if (fp < 0) {
		fprint(2, "couldn't open file...\n");
		exits("failure");
	}

  /* sanity check */
  if (numsamples > MAX_SAMPLES) {
    fprint(2,"WARNING: sample count exceeds maximum.\n");
    fprint(2,"         setting count to maximum.\n");
    numsamples = MAX_SAMPLES;
  }
  
  if (interval_bits > MAX_BITS || interval_bits < MIN_BITS) {
    fprint(2,"WARNING: interval bits invalid.  set to %d.\n",
	    MAX_BITS);
    interval_bits = MAX_BITS;
  }

  /* set up sampling.  first, take a few bogus samples to warm up the
     cache and pipeline */
  interval_length = 1 << interval_bits;  
  done = 0;
  count = 0;

  samples[done*2] = 0;
  cycles(&last);
  endinterval = (last + interval_length) & (~(interval_length - 1));

#ifdef _WITH_TAU_
  TAU_PROFILE_START(t1);
#endif /* _WITH_TAU_ */

  while (1) {
    count = 0;

#ifdef _WITH_TAU_
    TAU_PROFILE_START(looptimer);
#endif /* _WITH_TAU_ */

    for (now = last; now < endinterval; ) {
#ifdef MULTIITER
      for (k=0;k<ITERCOUNT;k++)
	count++;
      for (k=0;k<(ITERCOUNT-1);k++)
	count--;
#else
      count++;
#endif

     cycles(&now);
    }
#ifdef _WITH_TAU_
    TAU_PROFILE_STOP(looptimer);
#endif /* _WITH_TAU_ */

#ifdef _WITH_TAU_
    TAU_PROFILE_START(datasave);
#endif /* _WITH_TAU_ */
    samples[done*2] = last;
    samples[done*2 + 1] = count;
#ifdef _WITH_TAU_
    TAU_PROFILE_STOP(datasave);
#endif /* _WITH_TAU_ */

    done++;

    if (done > numsamples)
      break;

    cycles(&last);

    endinterval = (last + interval_length) & (~(interval_length - 1));
  }
#ifdef _WITH_TAU_
  TAU_PROFILE_STOP(t1);
#endif /* _WITH_TAU_ */

#ifdef _WITH_TAU_
  TAU_PROFILE_START(io);
#endif /* _WITH_TAU_ */
  if (use_stdout == 1) {
    for (i=0;i<numsamples;i++) {
      fprint(1,"%llud %llud\n",samples[i*2],samples[i*2 + 1]);
    }
  } else {
    for (i=0;i<numsamples;i++) {
      fprint(fp,"%llud\n",samples[i*2]);
    }
    close(fp);
    
    fp = create(fname_counts,OWRITE,0666);
    for (i=0;i<numsamples;i++) {
      fprint(fp,"%llud\n",samples[i*2 + 1]);
    }
    close(fp);
  }
#ifdef _WITH_TAU_
  TAU_PROFILE_STOP(io);
#endif /* _WITH_TAU_ */

#ifdef _WITH_TAU_
  TAU_PROFILE_STOP(mtimer);
#endif
  exits(nil);
}
